/*
 *  Copyright (C) 2010-2013 Axel Morgner, structr <structr@structr.org>
 *
 *  This file is part of structr <http://structr.org>.
 *
 *  structr is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  structr is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with structr.  If not, see <http://www.gnu.org/licenses/>.
 */

var StructrModel = {
    objects: {},
    callbacks: [],
    obj: function(id) {
        return StructrModel.objects[id];
    },
    createSearchResult: function(data) {

        var obj = new StructrSearchResult(data);

        // Store a reference of this object
        StructrModel.objects[data.id] = obj;

        // Check if the object is already contained in page
        var el = $('#id_' + obj.id);
        if (el && el.length) {
            return obj;
        }

        StructrModel.append(obj);

        return obj;

    },
    /**
     * Create a new object in the model and potentially append a UI element
     * If refId is set, insert before this node
     */
    create: function(data, refId, append) {

        log("StructrModel.create", data);
        if (!data) return;

        var type = data.type;
        
        var obj;
        
        if (type === 'Page') {

            obj = new StructrPage(data);

        } else if (type === 'Widget') {

            obj = new StructrWidget(data);

        } else if (type === 'Content') {

            obj = new StructrContent(data);

        } else if (type === 'Group') {

            obj = new StructrGroup(data);

        } else if (type === 'User') {

            obj = new StructrUser(data);

        } else if (type === 'File') {

            obj = new StructrFile(data);

        } else if (type === 'Image') {

            obj = new StructrImage(data);

        } else if (type === 'Folder') {

            obj = new StructrFolder(data);

        } else if (type === 'DataNode') {

            obj = new StructrDataNode(data);

        } else if (type === 'PropertyDefinition') {

            obj = new StructrPropertyDefinition(data);

        } else {

            obj = new StructrElement(data);

        }
        
        // Store a reference of this object
        StructrModel.objects[data.id] = obj;

        // Check if the object is already contained in page
        if (obj) {

            if (obj.exists && obj.exists()) {
                var el = Structr.node(obj.id);
                if (el.parent().prop('id') === 'elementsArea') {
                    el.remove();
                } else {
                    log('obj exists');
                    return obj;
                }
            }

        }

        if (refId || append === undefined || append) {
            StructrModel.append(obj, refId);
        }

        return obj;

    },
    /**
     * Append and check expand status
     */
    append: function(obj, refId) {

        var refNode = refId ? Structr.node(refId) : undefined;

        // Display in page (before refNode, if given)
        obj.append(refNode);

    },
    /**
     * Check expand status
     */
    expand: function(element, obj) {

        log('StructrModel.expand', element, obj);

        if (element) {

            if (isExpanded(obj.id)) {
                _Entities.ensureExpanded(element);
            }

            var parent = element.parent();

            if (parent && parent.hasClass('node') && parent.children('.node') && parent.children('.node').length) {

                log('parent of last appended object has children');

                var ent = Structr.entityFromElement(parent);
                _Entities.ensureExpanded(parent);
                log('entity', ent);
                _Entities.appendExpandIcon(parent, ent, true, true);

            }
        }
    },
    /**
     * Deletes an object
     */
    del: function(id) {

        var node = Structr.node(id);
        if (node) {
            node.remove();
        }
        removeExpandedNode(id);
        $('#show_' + id, previews).remove();
        _Pages.reloadPreviews();
        if (graph) {
            graph.redrawRelationships();
        }

    },
    /**
     * Update the model with the given data.
     * 
     * This function is usually triggered by a websocket message
     * and will trigger a UI refresh.
     **/
    update: function(data) {
        var obj = StructrModel.obj(data.id);

        if (obj) {

            $.each(Object.keys(data.data), function(i, key) {
                log('update model', key, data.data[key]);
                obj[key] = data.data[key];
                //console.log('object ', obj, 'updated with key', key, '=', obj[key]);
                //StructrModel.refreshKey(obj.id, key);
            });

            StructrModel.refresh(obj.id);

        }

        return obj;

    },
    updateKey: function(id, key, value) {
        log('StructrModel.updateKey', id, key, value);
        var obj = StructrModel.obj(id);

        if (obj) {
            obj[key] = value;
        }

        StructrModel.refreshKey(id, key);

    },
    /**
     * Refresh the object's UI representation with
     * the current model value for the given key
     */
    refreshKey: function(id, key) {

        var obj = StructrModel.obj(id);
        if (!obj)
            return;

        var element = Structr.node(id);

        if (!element)
            return;

        //for (var key in data.data) {
        var inputElement = $('td.' + key + '_ input', element);
        log(inputElement);
        var newValue = obj[key];
        //console.log(key, newValue, typeof newValue);

        var attrElement = element.children('.' + key + '_');

        if (attrElement && $(attrElement).length) {
            var tag = $(attrElement).get(0).tagName.toLowerCase();

            if (typeof newValue === 'boolean') {

                _Entities.changeBooleanAttribute(attrElement, newValue);

            } else {

                blinkGreen(attrElement);

                if (attrElement && tag === 'select') {
                    attrElement.val(newValue);
                } else {
                    log(key, newValue);
                    if (key === 'name') {
                        attrElement.html(fitStringToSize(newValue, 200));
                        attrElement.attr('title', newValue);
                    }
                }

                if (inputElement) {
                    inputElement.val(newValue);
                }

                if (key === 'content') {

                    log(attrElement.text(), newValue);

                    attrElement.text(newValue);

                    // hook for CodeMirror edit areas
                    //                        if (editor && editor.id == id) {
                    //                            log(editor.id);
                    //                            editor.setValue(newValue);
                    //                            editor.setCursor(editorCursor);
                    //                        }
                }
            }
        }

        log(key, Structr.getClass(element));

        if (key === 'name' && Structr.getClass(element) === 'page') {

            // update tab and reload iframe
            var tabNameElement = $('#show_' + id).children('.name_');

            blinkGreen(tabNameElement);

            tabNameElement.html(fitStringToSize(newValue, 200));
            tabNameElement.attr('title', newValue);

            log('Reload iframe', id, newValue);
            window.setTimeout(function() {
                _Pages.reloadIframe(id, newValue)
            }, 100);
        }

    },
    /**
     * Refresh the object's UI representation
     * with the current object data from the model.
     */
    refresh: function(id) {

        var obj = StructrModel.obj(id);

        if (obj) {
            var element = Structr.node(id);

            if (!element)
                return;

            log(obj, id, element);

            // update values with given key
            $.each(Object.keys(obj), function(i, key) {
                StructrModel.refreshKey(id, key);
            });

            // check display of HTML 'class' and 'id' attribute
            var id = obj._html_id ? obj._html_id.replace(/\${.*}/g, '${…}') : '';
            var cl = obj._html_class ? obj._html_class.replace(/\${.*}/g, '${…}') : '';

            var idEl = $(element).children('._html_id_');
            if (id) {
                if (!idEl.length) {
                    $('<span class="_html_id_"></span>').insertAfter($(element).children('b'));
                    clEl = $(element).children('._html_id_');
                }
                idEl.text('#' + id);
            } else {
                idEl.empty();
            }

            var clEl = $(element).children('._html_class_');
            if (cl) {
                if (!clEl.length) {
                    $('<span class="_html_class_"></span>').insertAfter(idEl.length ? idEl : $(element).children('b'));
                    clEl = $(element).children('._html_class_');
                }
                clEl.text('.' + $.trim(cl).replace(/ /g, '.'));
            } else {
                clEl.empty();
            }


            // check if key icon needs to be displayed (in case of nodes not visible to public/auth users)
            var protected = !obj.visibleToPublicUsers || !obj.visibleToAuthenticatedUsers;
            var keyIcon = $(element).children('.key_icon');
            if (!keyIcon.length) {
                // Images have a special subnode containing the icons
                keyIcon = $('.icons', $(element)).children('.key_icon');
            }
            if (protected) {
                keyIcon.show();
                keyIcon.addClass('donthide');
            } else {
                keyIcon.hide();
                keyIcon.removeClass('donthide');
            }

        }

    },
    /**
     * Fetch data from server. This will trigger a refresh of the model.
     */
    fetch: function(id) {
        Command.get(id);
    },
    /**
     * Save model data to server. This will trigger a refresh of the model.
     */
    save: function(id) {
        var obj = StructrModel.obj(id);
        log('StructrModel.save', obj);

        // Filter out object type data
        var data = {};
        $.each(Object.keys(obj), function(i, key) {

            var value = obj[key];

            if (typeof value !== 'object') {
                data[key] = value;
            }

        });
        //console.log('save', id, data);
        Command.setProperties(id, data);
    },
            
    callCallback: function(callback, entity) {

        log('Calling callback', callback, 'on entity', entity);

        if (callback && StructrModel.callbacks[callback]) {
            StructrModel.callbacks[callback](entity);
        }

    },
            
    clearCallback : function(callback) {

        removeFromArray(StructrModel.callbacks, callback);
        
    }

}


/**************************************
 * Structr Folder
 **************************************/

function StructrFolder(data) {
    var self = this;
    $.each(Object.keys(data), function(i, key) {
        self[key] = data[key];
    });
}

StructrFolder.prototype.save = function() {
    StructrModel.save(this.id);
}

StructrFolder.prototype.setProperty = function(key, value, recursive, callback) {
    Command.setProperty(this.id, key, value, recursive, callback);
}

StructrFolder.prototype.remove = function() {

    var folder = this;
    var parentFolder = StructrModel.obj(folder.parent.id);
    var parentFolderEl = Structr.node(parentFolder.id);

    if (!parentFolderEl)
        return;

    parentFolder.folders = removeFromArray(parentFolder.folders, folder);

    if (!parentFolder.files.length && !parentFolder.folders.length) {
        _Entities.removeExpandIcon(parentFolderEl);
        enable(parentFolderEl.children('.delete_icon')[0]);
    }

    var folderEl = Structr.node(folder.id);
    if (!folderEl)
        return;

    _Entities.resetMouseOverState(folderEl);

    folderEl.children('.delete_icon').replaceWith('<img title="Delete folder ' + folder.id + '" '
            + 'alt="Delete folder ' + folder.id + '" class="delete_icon button" src="' + Structr.delete_icon + '">');

    folderEl.children('.delete_icon').on('click', function(e) {
        e.stopPropagation();
        _Entities.deleteNode(this, folder);
    });

    folders.append(folderEl);

    folderEl.draggable({
        revert: 'invalid',
        containment: '#main',
        stack: '.folder'
    });
}

StructrFolder.prototype.append = function(refNode) {
    StructrModel.expand(_Files.appendFolderElement(this, refNode), this);
}


/**************************************
 * Structr File
 **************************************/

function StructrFile(data) {
    var self = this;
    $.each(Object.keys(data), function(i, key) {
        self[key] = data[key];
    });
}

StructrFile.prototype.save = function() {
    StructrModel.save(this.id);
}

StructrFile.prototype.setProperty = function(key, value, recursive, callback) {
    Command.setProperty(this.id, key, value, false, callback);
}

StructrFile.prototype.remove = function() {
    var file = this;
    var parentFolder = StructrModel.obj(file.parent.id);
    var parentFolderEl = Structr.node(parentFolder.id);

    parentFolder.files = removeFromArray(parentFolder.files, file);
    if (!parentFolder.files.length && !parentFolder.folders.length) {
        _Entities.removeExpandIcon(parentFolderEl);
        enable(parentFolderEl.children('.delete_icon')[0]);
    }

    var parentElement = files;


    var fileEl = Structr.node(file.id);
    if (!fileEl)
        return;

    _Entities.resetMouseOverState(fileEl);

    parentElement.append(fileEl);


    fileEl.children('.delete_icon').replaceWith('<img title="Delete File ' + file.id + '" '
            + 'alt="Delete File ' + file.id + '" class="delete_icon button" src="' + Structr.delete_icon + '">');

    fileEl.children('.delete_icon').on('click', function(e) {
        e.stopPropagation();
        _Entities.deleteNode(this, file);
    });

    fileEl.draggable({
        revert: 'invalid',
        containment: '#main',
        stack: 'div'
    });
}

StructrFile.prototype.append = function(refNode) {
    var file = this;
    if (refNode) {
        var parentFolder = StructrModel.obj(refNode.id);
        parentFolder.files.push(file);
    }
    StructrModel.expand(_Files.appendFileElement(this, refNode), this);
}


/**************************************
 * Structr Image
 **************************************/

function StructrImage(data) {
    var self = this;
    $.each(Object.keys(data), function(i, key) {
        self[key] = data[key];
    });
}

StructrImage.prototype.save = function() {
    StructrModel.save(this.id);
}

StructrImage.prototype.setProperty = function(key, value, recursive, callback) {
    Command.setProperty(this.id, key, value, false, callback);
}

StructrImage.prototype.remove = function() {
    _Images.removeImageFromFolder(this.id);
}

StructrImage.prototype.append = function(refNode) {
    StructrModel.expand(_Images.appendImageElement(this, refNode), this);
}


/**************************************
 * Structr User
 **************************************/

function StructrUser(data) {
    var self = this;
    $.each(Object.keys(data), function(i, key) {
        self[key] = data[key];
    });
}

StructrUser.prototype.save = function() {
    StructrModel.save(this.id);
}

StructrUser.prototype.setProperty = function(key, value, recursive, callback) {
    Command.setProperty(this.id, key, value, false, callback);
}

StructrUser.prototype.remove = function() {
    var user = this;
    var group = StructrModel.obj(user.groups[0]);
    var groupEl = Structr.node(group.id);

    group.members = removeFromArray(group.members, user);
    if (!group.members.length) {
        _Entities.removeExpandIcon(groupEl);
        enable(groupEl.children('.delete_icon')[0]);
    }

    var parentElement = users;

    var userEl = Structr.node(user.id);
    if (!userEl)
        return;

    _Entities.resetMouseOverState(userEl);

    parentElement.append(userEl);

    userEl.children('.delete_icon').replaceWith('<img title="Delete User ' + user.id + '" '
            + 'alt="Delete User ' + user.id + '" class="delete_icon button" src="' + Structr.delete_icon + '">');

    userEl.children('.delete_icon').on('click', function(e) {
        e.stopPropagation();
        _Entities.deleteNode(this, user);
    });

    userEl.draggable({
        revert: 'invalid',
        containment: '#main',
        stack: 'div'
    });
}

StructrUser.prototype.append = function(refNode) {
    var user = this;
    if (refNode) {
        var group = StructrModel.obj(refNode.id);
        group.members.push(user);
    }
    StructrModel.expand(_UsersAndGroups.appendUserElement(this, refNode), this);
}

/**************************************
 * Structr Group
 **************************************/

function StructrGroup(data) {
    var self = this;
    $.each(Object.keys(data), function(i, key) {
        self[key] = data[key];
    });
}

StructrGroup.prototype.save = function() {
    StructrModel.save(this.id);
}

StructrGroup.prototype.setProperty = function(key, value, recursive, callback) {
    Command.setProperty(this.id, key, value, recursive, callback);
}

StructrGroup.prototype.append = function(refNode) {
    StructrModel.expand(_UsersAndGroups.appendGroupElement(this, refNode), this);
}

/**************************************
 * Structr Page
 **************************************/

function StructrPage(data) {
    var self = this;
    $.each(Object.keys(data), function(i, key) {
        self[key] = data[key];
    });
}

//StructrPage.prototype.createElement = function(name) {
//    return new Element(name);
//}

StructrPage.prototype.setProperty = function(key, value, recursive, callback) {
    Command.setProperty(this.id, key, value, recursive, callback);
}

StructrPage.prototype.append = function() {
    StructrModel.expand(_Pages.appendPageElement(this), this);
}

/**************************************
 * Structr Widget
 **************************************/

function StructrWidget(data) {
    var self = this;
    $.each(Object.keys(data), function(i, key) {
        self[key] = data[key];
    });
}

//StructrPage.prototype.createElement = function(name) {
//    return new Element(name);
//}

StructrWidget.prototype.setProperty = function(key, value, recursive, callback) {
    Command.setProperty(this.id, key, value, recursive, callback);
}

StructrWidget.prototype.append = function() {
    StructrModel.expand(_Widgets.appendWidgetElement(this), this);
}

/**************************************
 * Structr Element
 **************************************/

function StructrElement(data) {
    var self = this;
    $.each(Object.keys(data), function(i, key) {
        self[key] = data[key];
    });
}

StructrElement.prototype.appendChild = function(el) {
    var self = this;
    self.children.push(el);
}

StructrElement.prototype.setProperty = function(key, value, recursive, callback) {
    Command.setProperty(this.id, key, value, recursive, callback);
}

StructrElement.prototype.removeAttribute = function(key) {
    var self = this;
    delete self[key];
}

StructrElement.prototype.save = function() {
    StructrModel.save(this.id);
}

StructrElement.prototype.remove = function() {
    var element = Structr.node(this.id);
    if (this.parent) {
        var parent = Structr.node(this.parent.id);
    }

    if (element) {
        // If element is removed from page tree, reload elements area
        if (element.closest('#pages').length) {
            _Elements.reloadUnattachedNodes();
        }
        element.remove();
    }

    log(this, element, parent, Structr.containsNodes(parent));

    if (element && parent && !Structr.containsNodes(parent)) {
        _Entities.removeExpandIcon(parent);
    }
    _Pages.reloadPreviews();
}

StructrElement.prototype.append = function(refNode) {
    StructrModel.expand(_Pages.appendElementElement(this, refNode), this);
}

StructrElement.prototype.exists = function() {

    var obj = this;

    var hasChildren = obj.childrenIds && obj.childrenIds.length;
    var isComponent = obj.syncedNodes && obj.syncedNodes.length;

    var isMasterComponent = (isComponent && hasChildren);

    return !isMasterComponent && Structr.node(obj.id);
}


/**************************************
 * Structr Content
 **************************************/

function StructrContent(data) {
    var self = this;
    $.each(Object.keys(data), function(i, key) {
        self[key] = data[key];
    });
}

StructrContent.prototype.appendChild = function(el) {
    var self = this;
    self.children.push(el);
}

StructrContent.prototype.setProperty = function(key, value, recursive, callback) {
    Command.setProperty(this.id, key, value, recursive, callback);
}

//StructrContent.prototype.setProperties = function(attributes) {
//    this.attributes = attributes;
//}

StructrContent.prototype.removeAttribute = function(key) {
    var self = this;
    delete self[key];
}

StructrContent.prototype.save = function() {
    StructrModel.save(this.id);
}

StructrContent.prototype.remove = function() {
    var element = Structr.node(this.id);
    if (this.parent) {
        var parent = Structr.node(this.parent.id);
    }

    if (element) {
        // If element is removed from page tree, reload elements area
        if (element.closest('#pages').length) {
            _Elements.reloadUnattachedNodes();
        }
        element.remove();
    }

    if (parent && !Structr.containsNodes(parent)) {
        _Entities.removeExpandIcon(parent);
    }
    _Pages.reloadPreviews();
}

StructrContent.prototype.append = function(refNode) {

    var id = this.id;
    var parentId;

    var parent;
    if (this.parent) {
        parentId = this.parent.id;
        parent = Structr.node(parentId);
    }

    var div = _Contents.appendContentElement(this, refNode);
    if (!div)
        return;

    log('appendContentElement div', div);

    StructrModel.expand(div, this);

    if (parent) {

        $('.button', div).on('mousedown', function(e) {
            e.stopPropagation();
        });

        $('.delete_icon', div).replaceWith('<img title="Remove content element from parent ' + parentId + '" '
                + 'alt="Remove content element from parent ' + parentId + '" class="delete_icon button" src="' + _Contents.delete_icon + '">');
        $('.delete_icon', div).on('click', function(e) {
            e.stopPropagation();
            Command.removeChild(id);
        });
    }

    _Entities.setMouseOver(div);

}
StructrContent.prototype.exists = function() {

    return Structr.node(this.id);
}

/**************************************
 * Structr PropertyDefinition
 **************************************/

function StructrPropertyDefinition(data) {
    var self = this;
    $.each(Object.keys(data), function(i, key) {
        self[key] = data[key];
    });
}

StructrPropertyDefinition.prototype.save = function() {
    StructrModel.save(this.id);
}

StructrPropertyDefinition.prototype.setProperty = function(key, value, recursive, callback) {
    Command.setProperty(this.id, key, value, false, callback);
}

StructrPropertyDefinition.prototype.remove = function() {
    alert('FIXME: Implement StructrPropertyDefinition.remove()!');
}

StructrPropertyDefinition.prototype.append = function(refNode) {
    //_Types.registerPropertyDefinitionElement(this);
}


/**************************************
 * Search result
 **************************************/

function StructrSearchResult(data) {
    var self = this;
    $.each(Object.keys(data), function(i, key) {
        self[key] = data[key];
    });
}

StructrSearchResult.prototype.save = function() {
    StructrModel.save(this.id);
}

StructrSearchResult.prototype.setProperty = function(key, value, recursive, callback) {
    Command.setProperty(this.id, key, value, recursive, callback);
}

StructrSearchResult.prototype.append = function() {
    _Dashboard.appendNode(this);
}



function removeFromArray(array, obj) {
    var newArray = [];
    $.each(array, function(i, el) {
        if (el.id !== obj.id) {
            newArray.push(el);
        }
    });
    return newArray;
}
